{********************************************************}
{                                                        }
{                 Zeos Database Objects                  }
{              Interbase direct class API                }
{                                                        }
{       Copyright (c) 1999-2001 Sergey Seroukhov         }
{                                                        }
{********************************************************}

unit ZDirIbSql;

interface

uses Windows, Classes, SysUtils, Db, ZLibIbSql, ZDirSql, ZSqlTypes, ZTransact;

{$INCLUDE ..\Zeos.inc}

type
  { Interbase Parameter Block class }
  TIbParamList = class
  private
    FList: TStringList;

    function  GetParam(Index: Word): string;
    procedure SetParam(Index: Word; Value: string);
  public
    constructor Create;
    destructor Destroy; override;

    function  IndexOf(Index: Word): Integer;
    procedure Add(Index: Word; Value: string);
    procedure Delete(Index: Word);
    procedure Clear;

    procedure GenerateDPB(var DPB: string; var DPBLength: Short);
    procedure GenerateTPB(var TPB: string; var TPBLength: Short);

    property Params[Index: Word]: string read GetParam write SetParam;
  end;

  { Interbase status array }
  ARRAY_ISC_STATUS = array[0.. 20] of ISC_STATUS;

  { Direct connection to Interbase database }
  TDirIbSqlConnect = class (TDirConnect)
  private
    FHandle: TISC_DB_HANDLE;
    FStatusVector: ARRAY_ISC_STATUS;
    FParams: TIbParamList;
    FDialect: Word;
  protected
    function GetErrorMsg: ShortString; override;
    function GetFullDbName: string;
    function CheckResult(Cmd: string): Boolean;
    function GetStatusVector(Index: Word): ISC_STATUS;
    procedure SetStatusVector(Index: Word; Value: ISC_STATUS);
  public
    constructor Create;
    destructor  Destroy; override;

    procedure Connect; override;
    procedure Disconnect; override;
    procedure CreateDatabase(Params: string); override;
    procedure DropDatabase; override;

    property Handle: TISC_DB_HANDLE read FHandle;
    property StatusVector[Index: Word]: ISC_STATUS read GetStatusVector
      write SetStatusVector;
    property Params: TIbParamList read FParams;
    property Dialect: Word read FDialect write FDialect;
  end;

  { Direct Interbase transaction }
  TDirIbSqlTransact = class (TDirTransact)
  private
    FHandle: TISC_TR_HANDLE;
    FParams: TIbParamList;
  public
    constructor Create(AConnect: TDirIbSqlConnect);
    destructor Destroy; override;

    procedure Open; override;
    procedure Close; override;
    procedure StartTransaction; override;
    procedure EndTransaction; override;
    procedure Commit; override;
    procedure Rollback; override;

    property Handle: TISC_TR_HANDLE read FHandle;
    property Params: TIbParamList read FParams;
  end;

{ Maximum xsqlvar buffer }
const MAX_XSQLVAR = 50;

type
  { Direct Interbase Query }
  TDirIbSqlQuery = class(TDirQuery)
  private
    FHandle: TISC_STMT_HANDLE;
    FOutSqlDa: PXSQLDA;
    FInSqlDa: PXSQLDA;

    function FreeStatement: Boolean;
  protected
    function GetErrorMsg: ShortString; override;
  public
    constructor Create(AConnect: TDirIbSqlConnect; ATransact: TDirIbSqlTransact);
    destructor  Destroy; override;

    function  Execute: LongInt; override;
    function  ExecuteParams(Params: TVarRecArray;
      ParamCount: Integer): LongInt; override;
    procedure Open; override;
    procedure Close; override;
    function  CreateBlobObject: TDirBlob; override;

    procedure ShowDatabases(DatabaseName: ShortString); override;
    procedure ShowTables(TableName: ShortString); override;
    procedure ShowColumns(TableName, ColumnName: ShortString); override;
    procedure ShowIndexes(TableName: ShortString); override;

    procedure First; override;
    procedure Last; override;
    procedure Prev; override;
    procedure Next; override;
    procedure Go(Num: Integer); override;

    function  FieldCount: Integer; override;
    function  RecordCount: Integer; override;

    function  FieldName(FieldNum: Integer): ShortString; override;
    function  FieldSize(FieldNum: Integer): Integer; override;
    function  FieldMaxSize(FieldNum: Integer): Integer; override;
    function  FieldDecimals(FieldNum: Integer): Integer; override;
    function  FieldType(FieldNum: Integer): Integer; override;
    function  FieldDataType(FieldNum: Integer): TFieldType; override;
    function  FieldIsNull(FieldNum: Integer): Boolean; override;
    function  Field(FieldNum: Integer): string; override;
    function  FieldBuffer(FieldNum: Integer): PChar; override;

    function StringToSql(Value: string): string; override;

    property Handle: TISC_STMT_HANDLE read FHandle;
  end;

  { Abstract class for database binary large object }
  TDirIbSqlBlob = class(TDirBlob)
  private
    FBlobHandle: TISC_BLOB_HANDLE;
  public
    constructor Create(AConnect: TDirConnect; ATransact: TDirTransact;
      AHandle: TBlobHandle);

    procedure Open(Mode: Integer); override;
    procedure Close; override;
    procedure CreateBlob; override;
    procedure DropBlob; override;

    function  Read(Buffer: PChar; Length: Integer): Integer; override;
    function  Write(Buffer: PChar; Length: Integer): Integer; override;

    property BlobHandle: TISC_BLOB_HANDLE read FBlobHandle;
  end;

{ Convert postgresql field types to delphi field types }
function IbSqlToDelphiType(Value, Precision: Integer): TFieldType;

{ Monitor list }
var MonitorList: TZMonitorList;

implementation

uses ZExtra, ZDBaseConst, Math, ZSqlExtra;

{*************** TIbParamList class implementation ***************}

{ Class constructor }
constructor TIbParamList.Create;
begin
  FList := TStringList.Create;
end;

{ Class destructor }
destructor TIbParamList.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

{ Get parameter }
function TIbParamList.GetParam(Index: Word): string;
var
  N: Integer;
begin
  N := IndexOf(Index);
  if N >= 0 then
    Result := FList[N]
  else
    Result := '';
end;

{ Set parameter }
procedure TIbParamList.SetParam(Index: Word; Value: string);
begin
  Add(Index, Value);
end;

{ Get param index }
function TIbParamList.IndexOf(Index: Word): Integer;
begin
  Result := FList.IndexOfObject(TObject(Index));
end;

{ Add new parameter }
procedure TIbParamList.Add(Index: Word; Value: string);
var
  N: Integer;
begin
  N := IndexOf(Index);
  if N >= 0 then
    FList[N] := Value
  else
    FList.AddObject(Value, TObject(Index));
end;

{ Delete parameter }
procedure TIbParamList.Delete(Index: Word);
var
  N: Integer;
begin
  N := IndexOf(Index);
  if N >= 0 then
    FList.Delete(N);
end;

{ Clear param list }
procedure TIbParamList.Clear;
begin
  FList.Clear;
end;

{ Fill database parameter block }
procedure TIbParamList.GenerateDPB(var DPB: string; var DPBLength: Short);
var
  I, PValue: Integer;
  ParamNo: Word;
  ParamValue: string;
begin
  DPBLength := 1;
  DPB := Char(isc_dpb_version1);

  for I := 0 to FList.Count - 1 do
  begin
    ParamNo := Word(FList.Objects[I]);
    ParamValue := FList[I];

    case ParamNo of
      isc_dpb_user_name, isc_dpb_password, isc_dpb_password_enc,
      isc_dpb_sys_user_name, isc_dpb_license, isc_dpb_encrypt_key,
      isc_dpb_lc_messages, isc_dpb_lc_ctype, isc_dpb_sql_role_name:
        begin
          DPB := DPB + Char(ParamNo) + Char(Length(ParamValue)) + ParamValue;
          Inc(DPBLength, 2 + Length(ParamValue));
        end;
      isc_dpb_num_buffers, isc_dpb_dbkey_scope, isc_dpb_force_write,
      isc_dpb_no_reserve, isc_dpb_damaged, isc_dpb_verify:
        begin
          DPB := DPB + Char(ParamNo) + #1 + Char(StrToInt(ParamValue));
          Inc(DPBLength, 3);
        end;
      isc_dpb_sweep:
        begin
          DPB := DPB + Char(ParamNo) + #1 + Char(isc_dpb_records);
          Inc(DPBLength, 3);
        end;
      isc_dpb_sweep_interval:
        begin
          PValue := StrToInt(ParamValue);
          DPB := DPB + Char(ParamNo) + #4 + PChar(@PValue)[0] + PChar(@PValue)[1] +
            PChar(@PValue)[2] + PChar(@PValue)[3];
          Inc(DPBLength, 6);
        end;
      isc_dpb_activate_shadow, isc_dpb_delete_shadow, isc_dpb_begin_log,
      isc_dpb_quit_log:
        begin
          DPB := DPB + Char(ParamNo) + #1 + #0;
          Inc(DPBLength, 3);
        end;
    end;
  end;
end;

{ Fill transaction parameter block }
procedure TIbParamList.GenerateTPB(var TPB: string; var TPBLength: Short);
var
  I: Integer;
  ParamNo: Word;
  ParamValue: string;
begin
  if FList.Count = 0 then
  begin
    TPB := '';
    TPBLength := 0;
    Exit;
  end
  else
  begin
    TPB := Char(isc_tpb_version3);
    TPBLength := 1;
  end;

  for I := 0 to FList.Count - 1 do
  begin
    ParamNo := Word(FList.Objects[I]);
    ParamValue := FList[I];

    case ParamNo of
      isc_tpb_consistency, isc_tpb_exclusive, isc_tpb_concurrency,
      isc_tpb_shared, isc_tpb_wait, isc_tpb_nowait, isc_tpb_read,
      isc_tpb_write, isc_tpb_ignore_limbo, isc_tpb_read_committed,
      isc_tpb_rec_version, isc_tpb_no_rec_version:
        begin
          TPB := TPB + Char(ParamNo);
          Inc(TPBLength, 1);
        end;
      isc_tpb_lock_read, isc_tpb_lock_write:
       begin
         TPB := TPB + Char(ParamNo) + Char(Length(ParamValue)) + ParamValue;
        Inc(TPBLength, Length(ParamValue) + 2);
      end;
    end;
  end;
end;

{**************** TDirIbSqlConnect implementation ************}

{ Class constructor }
constructor TDirIbSqlConnect.Create;
begin
  inherited Create;
  FParams := TIbParamList.Create;
  FHandle := nil;
  FDialect := 1;
end;

{ Class destructor }
destructor TDirIbSqlConnect.Destroy;
begin
  inherited Destroy;
  FParams.Free;
end;

{ Get status vector item }
function TDirIbSqlConnect.GetStatusVector(Index: Word): ISC_STATUS;
begin
  Result := FStatusVector[Index];
end;

{ Set status vector item }
procedure TDirIbSqlConnect.SetStatusVector(Index: Word; Value: ISC_STATUS);
begin
  FStatusVector[Index] := Value;
end;

{ Check result code }
function TDirIbSqlConnect.CheckResult(Cmd: string): Boolean;
begin
  if (StatusVector[0] = 1) and (StatusVector[1] > 0) then
    SetStatus(csFail)
  else
    SetStatus(csOk);
  Result := (Status = csOk);
  MonitorList.InvokeEvent(Cmd, Error, not Result);
end;

{ Get full database name with host name }
function TDirIbSqlConnect.GetFullDbName: string;
begin
  Result := Trim(Database);
  HostName := Trim(HostName);
  if HostName <> '' then
    Result := HostName + ':' + Result;
end;

{ Connect to existed database }
procedure TDirIbSqlConnect.Connect;
var
  Temp: string;
  DBName: array[0..512] of Char;
  DPB: PChar;
  DPBLen: Short;
begin
  inherited Connect;
  if hDll = 0 then IbSqlLoadLib;

  if Login <> '' then
    Params.Add(isc_dpb_user_name, Login);
  if Passwd <> '' then
    Params.Add(isc_dpb_password, Passwd);
  Params.GenerateDPB(Temp, DPBLen);
  DPBLen := Length(Temp);
  DPB := StrAlloc(DPBLen + 1);
  StrPCopy(DPB, Temp);
  StrPCopy(DBName, GetFullDbName);

{ Connect database }
  FHandle := nil;
  isc_attach_database(@FStatusVector, StrLen(DBName), DBName, @Handle,
    DPBLen, DPB);
  StrDispose(DPB);

  SetActive(CheckResult(Format('CONNECT ''%s''',[GetFullDbName])));
end;

{ Disconnect from database }
procedure TDirIbSqlConnect.Disconnect;
begin
  if Active then
  begin
    isc_detach_database(@FStatusVector, @Handle);
    CheckResult(Format('DISCONNECT ''%s''',[GetFullDbName]));
    SetActive(False);
  end;
end;

{ Create new database }
procedure TDirIbSqlConnect.CreateDatabase(Params: string);
var
  TrHandle: TISC_TR_HANDLE;
  Buffer: string;
begin
  inherited CreateDatabase(Params);
  if hDll = 0 then IbSqlLoadLib;
  Disconnect;
  TrHandle := nil;

  Buffer := Format('CREATE DATABASE ''%s'' USER ''%s'' PASSWORD ''%s'' %s',
    [GetFullDbName, Login, Passwd, Params]);
  isc_dsql_execute_immediate(@FStatusVector, @Handle, @TrHandle, 0,
    PChar(Buffer), Dialect, nil);
  CheckResult(Buffer);
end;

{ Drop existed database }
procedure TDirIbSqlConnect.DropDatabase;
begin
  if not Active then Connect;
  if Active then
  begin
    isc_drop_database(@FStatusVector, @Handle);
    CheckResult(Format('DROP DATABASE ''%s''',[GetFullDbName]));
    SetActive(False);
  end;
end;

{ Get Interbase Database Error }
function TDirIbSqlConnect.GetErrorMsg: ShortString;
var
  PStatusVector: PISC_STATUS;
  Msg: array[0..1024] of Char;
begin
  if (FStatusVector[0] = 1) and (FStatusVector[1] > 0) then
  begin
    PStatusVector := @FStatusVector;
    isc_interprete(Msg, @PStatusVector);
    Result := StrPas(Msg);
  end else if not Active then
    Result := SNotConnected
  else Result := '';
end;

{************* TDirIbSqlTransact implementation *************}

{ Class constructor }
constructor TDirIbSqlTransact.Create(AConnect: TDirIbSqlConnect);
begin
  inherited Create;
  FParams := TIbParamList.Create;
  FHandle := nil;
  Connect := AConnect;
end;

{ Class destructor }
destructor TDirIbSqlTransact.Destroy;
begin
  inherited Destroy;
  FParams.Free;
end;

{ Connect transaction }
procedure TDirIbSqlTransact.Open;
begin
  inherited Open;
  SetStatus(csFail);
  StartTransaction;
  SetActive(Status = csOk);
end;

{ Disconnect transaction }
procedure TDirIbSqlTransact.Close;
begin
  EndTransaction;
  SetActive(False);
end;

{ Start transaction }
procedure TDirIbSqlTransact.StartTransaction;
var
  Temp: string;
  TPB: PChar;
  TPBLen: Short;
begin
  SetStatus(csFail);
  if not Assigned(Connect) or not Connect.Active then
    Exit;

  Params.GenerateTPB(Temp, TPBLen);
  TPBLen := Length(Temp);
  if TPBLen <> 0 then
  begin
    TPB := StrAlloc(TPBLen + 1);
    StrPCopy(TPB, Temp);
  end else
    TPB := nil;

  FHandle := nil;
  isc_start_transaction(@TDirIbSqlConnect(Connect).FStatusVector, @Handle, 1,
    @TDirIbSqlConnect(Connect).Handle, TPBLen, TPB);
  StrDispose(TPB);
  TDirIbSqlConnect(Connect).CheckResult('START TRANSACTION');
  SetStatus(Connect.Status);
end;

{ End transaction }
procedure TDirIbSqlTransact.EndTransaction;
begin
  SetStatus(csFail);
  if Active then
 begin
    isc_rollback_transaction(@TDirIbSqlConnect(Connect).FStatusVector, @Handle);
    TDirIbSqlConnect(Connect).CheckResult('END TRANSACTION');
    SetStatus(Connect.Status);
    FHandle := nil;
  end else
    SetStatus(csOk);
end;

{ Commit transaction }
procedure TDirIbSqlTransact.Commit;
begin
  SetStatus(csFail);
  if Active then
  begin
    isc_commit_retaining(@TDirIbSqlConnect(Connect).FStatusVector, @Handle);
    TDirIbSqlConnect(Connect).CheckResult('COMMIT');
    SetStatus(Connect.Status);
  end;
end;

{ Rollback transaction }
procedure TDirIbSqlTransact.Rollback;
begin
  SetStatus(csFail);
  if Active then
  begin
    isc_rollback_transaction(@TDirIbSqlConnect(Connect).FStatusVector, @Handle);
    TDirIbSqlConnect(Connect).CheckResult('ROLLBACK');
    SetStatus(Connect.Status);
    if Connect.Status = csOk then
      StartTransaction;
  end;
end;

{************* TDirIbSqlQuery implementation ************}

{ Class constructor }
constructor TDirIbSqlQuery.Create(AConnect: TDirIbSqlConnect;
  ATransact: TDirIbSqlTransact);
begin
  inherited Create;
  Connect := AConnect;
  Transact := ATransact;
  GetMem(FOutSqlDa, XSQLDA_LENGTH(MAX_XSQLVAR));
  FOutSqlDa.version := SQLDA_VERSION1;
  FOutSqlDa.sqln := MAX_XSQLVAR;

  GetMem(FInSqlDa, XSQLDA_LENGTH(MAX_XSQLVAR));
  FInSqlDa.version := SQLDA_VERSION1;
  FInSqlDa.sqln := MAX_XSQLVAR;
end;

{ Class destructor }
destructor TDirIbSqlQuery.Destroy;
begin
  inherited Destroy;
  FreeMem(FInSqlDa);
  FreeMem(FOutSqlDa);
end;

{ Get an error message }
function TDirIbSqlQuery.GetErrorMsg: ShortString;
var
  Msg: array[0..1024] of Char;
  SqlCode: LongInt;
  IbConnect: TDirIbSqlConnect;
begin
  Result := '';
  if not Assigned(Connect) then Exit;
  IbConnect := TDirIbSqlConnect(Connect);

  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    SqlCode := isc_sqlcode(@IbConnect.FStatusVector);
    isc_sql_interprete(SqlCode, Msg, 1024);
    Result := StrPas(Msg);
  end;
end;

{ Execute an SQL statement }
function TDirIbSqlQuery.Execute: LongInt;
begin
  inherited Execute;

  SetStatus(qsFail);
  Result := 0;
  if Assigned(Connect) and Assigned(Transact) then
  begin
    isc_dsql_execute_immediate(@TDirIbSqlConnect(Connect).FStatusVector,
      @TDirIbSqlConnect(Connect).Handle, @TDirIbSqlTransact(Transact).Handle,
{$IFDEF DELETE_QUERY_SPACES}
      0, PChar(ClearSpaces(Sql)), TDirIbSqlConnect(Connect).Dialect, nil);
{$ELSE}
      0, PChar(Sql), TDirIbSqlConnect(Connect).Dialect, nil);
{$ENDIF}
    TDirIbSqlConnect(Connect).CheckResult(Sql);
    if Connect.Status = csOk then
      SetStatus(qsCommandOk);
  end;
end;

{ Execute an SQL statement with params }
function TDirIbSqlQuery.ExecuteParams(Params: TVarRecArray; ParamCount: Integer): LongInt;
var
  I, NullFlag: Short;
  IbConnect: TDirIbSqlConnect;
  IbTransact: TDirIbSqlTransact;
  SqlVar: PXSQLVAR;
begin
  { Check connect and transaction status }
  Result := inherited Execute;
  SetStatus(qsFail);
  if not Assigned(Connect) or not Assigned(Transact)
    or not (Connect.Active and Transact.Active) then
  begin
    TDirIbSqlConnect(Connect).CheckResult(Sql);
    Exit;
  end;

  IbConnect := TDirIbSqlConnect(Connect);
  IbTransact := TDirIbSqlTransact(Transact);

  { Allocate an sql statement }
  FHandle := nil;
  isc_dsql_allocate_statement(@IbConnect.FStatusVector, @IbConnect.Handle, @FHandle);
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    IbConnect.CheckResult(Sql);
    Exit;
  end;

  { Prepare an sql statement }
  isc_dsql_prepare(@IbConnect.FStatusVector, @IbTransact.Handle, @FHandle, 0,
{$IFDEF DELETE_QUERY_SPACES}
    PChar(ClearSpaces(Sql)), IbConnect.Dialect, nil);
{$ELSE}
    PChar(Sql), IbConnect.Dialect, nil);
{$ENDIF}
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    IbConnect.CheckResult(Sql);
    Exit;
  end;

  isc_dsql_describe_bind(@IbConnect.FStatusVector, @FHandle, IbConnect.Dialect,
    FInSqlDa);
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    IbConnect.CheckResult(Sql);
    Exit;
  end;

  { Resize XSQLDA structure if needed }
  if FInSqlDa.sqld > FInSqlDa.sqln then
  begin
    ReallocMem(FInSqlDa, XSQLDA_LENGTH(FInSqlDa.sqld));
    FInSqlDa.sqln := FInSqlDa.sqld;
    isc_dsql_describe_bind(@IbConnect.FStatusVector, @FHandle, IbConnect.Dialect,
      FInSqlDa);
    if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
    begin
      IbConnect.CheckResult(Sql);
      Exit;
    end;
  end;

  { Allocate memory for result set }
  NullFlag := -1;
  for I := 0 to FInSqlDa.sqld-1 do
  begin
    SqlVar := @FInSqlDa.SqlVar[I];
    if (I > ParamCount) or (I > High(Params)) then
    begin
      SqlVar.sqldata := nil;
      SqlVar.sqltype := SqlVar.sqltype or 1;
      SqlVar.sqlind := @NullFlag;
    end
    else
    begin
      SqlVar.sqltype := SqlVar.sqltype and (not 1);
      case SqlVar.sqltype of
        SQL_VARYING, SQL_TEXT:
          begin
            SqlVar.sqldata := @Params[I];
//            SqlVar.sqllen := Length(string(Params[I]));
          end
        else
          SqlVar.sqldata := @Params[I];
      end;
      SqlVar.sqlind := nil;
    end;
  end;

  { Execute a query }
  isc_dsql_execute(@IbConnect.FStatusVector, @IbTransact.Handle, @FHandle,
    IbConnect.Dialect, FInSqlDa);
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
    SetStatus(qsFail)
  else
    SetStatus(qsCommandOk);
  FreeStatement;
  IbConnect.CheckResult(Sql);
end;

{ Free allocated statement }
function TDirIbSqlQuery.FreeStatement: Boolean;
var
  IbConnect: TDirIbSqlConnect;
begin
  Result := True;
  if FHandle <> nil then
  begin
    IbConnect := TDirIbSqlConnect(Connect);
    isc_dsql_free_statement(@IbConnect.FStatusVector, @FHandle, DSQL_CLOSE);
    if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
      Result := False;
  end;
end;

{ Open a sql query with result set }
procedure TDirIbSqlQuery.Open;
var
  I: Integer;
  IbConnect: TDirIbSqlConnect;
  IbTransact: TDirIbSqlTransact;
  SqlVar: PXSQLVAR;
begin
  inherited Open;

  { Check connect and transaction status }
  SetStatus(qsFail);
  if not Assigned(Connect) or not Assigned(Transact)
    or not (Connect.Active and Transact.Active) then
  begin
    TDirIbSqlConnect(Connect).CheckResult(Sql);
    Exit;
  end;

  IbConnect := TDirIbSqlConnect(Connect);
  IbTransact := TDirIbSqlTransact(Transact);

  { Allocate an sql statement }
  FHandle := nil;
  isc_dsql_allocate_statement(@IbConnect.FStatusVector, @IbConnect.Handle, @FHandle);
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    IbConnect.CheckResult(Sql);
    Exit;
  end;

  { Prepare an sql statement }
  isc_dsql_prepare(@IbConnect.FStatusVector, @IbTransact.Handle, @FHandle, 0,
{$IFDEF DELETE_QUERY_SPACES}
    PChar(ClearSpaces(Sql)), IbConnect.Dialect, FOutSqlDa);
{$ELSE}
    PChar(Sql), IbConnect.Dialect, FOutSqlDa);
{$ENDIF}
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    IbConnect.CheckResult(Sql);
    Exit;
  end;

  { Resize XSQLDA structure if needed }
  if FOutSqlDa.sqld > FOutSqlDa.sqln then
  begin
    ReallocMem(FOutSqlDa, XSQLDA_LENGTH(FOutSqlDa.sqld));
    FOutSqlDa.sqln := FOutSqlDa.sqld;
    isc_dsql_describe(@IbConnect.FStatusVector, @FHandle, IbConnect.Dialect,
      FOutSqlDa);
    if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
    begin
      IbConnect.CheckResult(Sql);
      Exit;
    end;
  end;

  { Allocate memory for result set }
  for I := 0 to FOutSqlDa.sqld-1 do
  begin
    SqlVar := @FOutSqlDa.SqlVar[I];
    case (SqlVar.sqltype and (not 1)) of
      SQL_VARYING:
        GetMem(SqlVar.sqldata, SqlVar.sqllen+2);
      SQL_TEXT:
        GetMem(SqlVar.sqldata, SqlVar.sqllen);
      SQL_LONG:
        GetMem(SqlVar.sqldata, SizeOf(LongInt));
      SQL_SHORT:
        GetMem(SqlVar.sqldata, SizeOf(Short));
      SQL_DOUBLE:
        GetMem(SqlVar.sqldata, SizeOf(Double));
      SQL_FLOAT:
        GetMem(SqlVar.sqldata, SizeOf(Single));
      SQL_ARRAY, SQL_BLOB, SQL_DATE, SQL_TYPE_DATE, SQL_TYPE_TIME:
        GetMem(SqlVar.sqldata, SizeOf(TISC_QUAD));
    end;
    if (SqlVar.sqltype and 1) <> 0 then
      GetMem(SqlVar.sqlind, SizeOf(Short))
    else
      SqlVar.sqlind := nil;
  end;

  { Execute a query }
  isc_dsql_execute(@IbConnect.FStatusVector, @IbTransact.Handle, @FHandle,
    IbConnect.Dialect, nil);
  if (IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0) then
  begin
    Close;
    SetStatus(qsFail);
    IbConnect.CheckResult(Sql);
    Exit;
  end;

  IbConnect.CheckResult(Sql);
  SetActive(True);
  SetStatus(qsTuplesOk);
  SetBOF(False);
  SetEOF(False);
  Next;
  if Status <> qsTuplesOk then
    SetActive(False);
end;

{ Close a sql query with result set }
procedure TDirIbSqlQuery.Close;
var
  I: Integer;
  SqlVar: PXSQLVAR;
begin
  if not Active then
    Exit;
  inherited Close;

  { Check connect and transaction status }
  SetStatus(qsTuplesOk);
  if not Assigned(Connect) then
    Exit;

  { Free sql statement }
  FreeStatement;

  { Free allocated memory }
  for I := 0 to FOutSqlDa.sqld-1 do
  begin
    SqlVar := @FOutSqlDa.SqlVar[I];
    FreeMem(SqlVar.sqldata);
    FreeMem(SqlVar.sqlind);
  end;
  FOutSqlDa.sqld := 0;
end;

{ Create linked blob object }
function TDirIbSqlQuery.CreateBlobObject: TDirBlob;
var
  TempHandle: TBlobHandle;
begin
  FillChar(TempHandle, SizeOf(TBlobHandle), 0);
  Result := TDirIbSqlBlob.Create(Connect, Transact, TempHandle);
end;

function TDirIbSqlQuery.Field(FieldNum: Integer): string;
var
  SqlVar: PXSQLVAR;
  VarType: Short;
  OldSep: Char;
  TempDate: TCTimeStructure;
begin
  Result := '';
  if not Active then Exit;
  SqlVar := @FOutSqlDa.SqlVar[FieldNum];
  VarType := SqlVar.sqltype;
  if ((VarType and 1) <> 0) and ((SqlVar.sqlind^) = -1) then
    Exit;
  OldSep := DecimalSeparator;
  DecimalSeparator := '.';
  case (VarType and (not 1)) of
    SQL_VARYING:
      Result := MemPas(SqlVar.sqldata+2, PSmallInt(SqlVar.sqldata)^);
    SQL_TEXT:
      Result := TrimRight(MemPas(SqlVar.sqldata, SqlVar.sqllen));
    SQL_LONG:
      begin
        if SqlVar.sqlscale = 0 then
          Result := IntToStr(PLongInt(SqlVar.sqldata)^)
        else
          Result := FloatToStr(PLongInt(SqlVar.sqldata)^
            / IntPower(10, SqlVar.sqlscale));
      end;
    SQL_SHORT:
      Result := IntToStr(PSmallInt(SqlVar.sqldata)^);
    SQL_DOUBLE:
      Result := FloatToStr(PDouble(SqlVar.sqldata)^);
    SQL_FLOAT:
      Result := FloatToStr(PSingle(SqlVar.sqldata)^);
    SQL_DATE:
      begin
        isc_decode_date(PISC_QUAD(SqlVar.sqldata), @TempDate);
        Result := DateTimeToSqlDate(EncodeDate(TempDate.tm_year + 1900,
          TempDate.tm_mon+1, TempDate.tm_mday) + EncodeTime(TempDate.tm_hour,
          TempDate.tm_min, TempDate.tm_sec, 0));
      end;
    SQL_BLOB:
      begin
        with TDirIbSqlBlob.Create(Connect, Transact,
          PBlobHandle(SqlVar.sqldata)^) do
        try
          Result := Value;
        finally
          Free;
        end;
      end;
  end;
  DecimalSeparator := OldSep;
end;

{ Check if field is null }
function TDirIbSqlQuery.FieldIsNull(FieldNum: Integer): Boolean;
var
  SqlVar: PXSQLVAR;
  VarType: Short;
begin
  Result := True;
  SqlVar := @FOutSqlDa.SqlVar[FieldNum];
  if not Active then Exit;
  VarType := SqlVar.sqltype;
  Result := ((VarType and 1) <> 0) and (SqlVar.sqlind <> nil)
    and ((SqlVar.sqlind^) = -1);
end;

{ Get field buffer }
function TDirIbSqlQuery.FieldBuffer(FieldNum: Integer): PChar;
var
  SqlVar: PXSQLVAR;
begin
  Result := nil;
  if not Active then Exit;
  SqlVar := @FOutSqlDa.SqlVar[FieldNum];
  if ((SqlVar.sqltype and 1) = 0) or ((SqlVar.sqlind^) <> -1) then
    Result := SqlVar.sqldata;
end;

{ Get field type }
function TDirIbSqlQuery.FieldType(FieldNum: Integer): Integer;
begin
  Result := 0;
  if not Active then Exit;
  Result := FOutSqlDa.SqlVar[FieldNum].sqltype and (not 1);
end;

{ Get field delphi compatible type }
function TDirIbSqlQuery.FieldDataType(FieldNum: Integer): TFieldType;
begin
  Result := ftUnknown;
  if not Active then Exit;
  case (FOutSqlDa.SqlVar[FieldNum].sqltype and (not 1)) of
    SQL_VARYING, SQL_TEXT: Result := ftString;
    SQL_LONG:
      if FOutSqlDa.SqlVar[FieldNum].sqlscale <> 0 then
        Result := ftFloat
      else
        Result := ftInteger;
    SQL_SHORT: Result := ftSmallInt;
    SQL_DOUBLE, SQL_FLOAT: Result := ftFloat;
    SQL_DATE: Result := ftDateTime;
    SQL_BLOB: Result := ftBlob;
    else
      Result := ftString;
  end;
end;

{ Get field count }
function TDirIbSqlQuery.FieldCount: Integer;
begin
  Result := 0;
  if not Active then Exit;
  Result := FOutSqlDa.sqld;
end;

{ Get field maximum size }
function TDirIbSqlQuery.FieldMaxSize(FieldNum: Integer): Integer;
begin
  Result := 0;
  if not Active then Exit;
  Result := FOutSqlDa.SqlVar[FieldNum].sqllen;
end;

{ Get fields decimals }
function TDirIbSqlQuery.FieldDecimals(FieldNum: Integer): Integer;
begin
  Result := 0;
  if not Active then Exit;
  Result := -FOutSqlDa.SqlVar[FieldNum].sqlscale;
end;

{ Get field name }
function TDirIbSqlQuery.FieldName(FieldNum: Integer): ShortString;
begin
  Result := '';
  if not Active then Exit;
  Result := MemPas(FOutSqlDa.SqlVar[FieldNum].aliasname,
    FOutSqlDa.SqlVar[FieldNum].aliasname_length);
  if Result = '' then Result := 'Field' + IntToStr(FieldNum+1);
end;

{ Get field size }
function TDirIbSqlQuery.FieldSize(FieldNum: Integer): Integer;
begin
  Result := 0;
  if not Active then Exit;
  Result := FOutSqlDa.SqlVar[FieldNum].sqllen;
  if (FOutSqlDa.SqlVar[FieldNum].sqltype and (not 1)) = SQL_VARYING then
    Result := PSmallInt(FOutSqlDa.SqlVar[FieldNum].sqldata)^;
end;

{ Go to the first row }
procedure TDirIbSqlQuery.First;
begin
end;

{ Go to specified row }
procedure TDirIbSqlQuery.Go(Num: Integer);
begin
end;

{ Go to the last row }
procedure TDirIbSqlQuery.Last;
begin
end;

{ Go to next row }
procedure TDirIbSqlQuery.Next;
var
  FetchStat: Integer;
  IbConnect: TDirIbSqlConnect;
begin
  if not Active or EOF then Exit;
  SetStatus(qsFail);
  if not Assigned(Connect) or not Assigned(Transact) then
    Exit;

  IbConnect := TDirIbSqlConnect(Connect);
  FetchStat := isc_dsql_fetch(@IbConnect.FStatusVector, @FHandle,
    IbConnect.Dialect, FOutSqlDa);
  if not ((IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0)) then
  begin
    SetStatus(qsTuplesOk);
    SetEOF(FetchStat <> 0);
    if FetchStat = 0 then
      SetRecNo(RecNo + 1);
  end else
    SetEOF(True);
end;

{ Go to prior row }
procedure TDirIbSqlQuery.Prev;
begin
end;

{ Get rows number }
function TDirIbSqlQuery.RecordCount: Integer;
var
  IbConnect: TDirIbSqlConnect;
  InBuf: Char;
  OutBuf: array[0..10] of Char;
begin
  Result := 0;
  if Active then
  begin
    IbConnect := TDirIbSqlConnect(Connect);
    InBuf := Chr(isc_info_sql_records);
    isc_dsql_sql_info(@IbConnect.FStatusVector, @FHandle, 1, @InBuf,
      SizeOf(OutBuf), OutBuf);
    if OutBuf[0] = Chr(isc_info_sql_records) then
      Result := isc_vax_integer(OutBuf+1, 4)
    else
      Result := RecNo;
  end;
end;

{ Showes table columns }
procedure TDirIbSqlQuery.ShowColumns(TableName, ColumnName: ShortString);
begin
  if Active then Close;
  SQL := 'SELECT RDB$FIELD_POSITION AS Idx, A.RDB$FIELD_NAME AS Fld,'
    +' B.RDB$FIELD_TYPE AS Typ, B.RDB$FIELD_LENGTH AS Len,'
    +' A.RDB$NULL_FLAG AS Nul, A.RDB$DEFAULT_SOURCE AS Def,'
    +' -B.RDB$FIELD_SCALE AS Scale'
    +' FROM RDB$RELATION_FIELDS A LEFT JOIN RDB$FIELDS B'
    +' ON A.RDB$FIELD_SOURCE=B.RDB$FIELD_NAME WHERE'
    +' A.RDB$RELATION_NAME='''+UpperCase(TableName)+'''';
  if ColumnName <> '' then
    SQL := SQL + ' AND A.RDB$FIELD_NAME LIKE '''+UpperCase(ColumnName)+'''';
  SQL := SQL + ' ORDER BY RDB$FIELD_POSITION';
  Open;
end;

{ Show existed databases }
procedure TDirIbSqlQuery.ShowDatabases(DatabaseName: ShortString);
begin
  inherited;
end;

{ Showes tables indices of database }
procedure TDirIbSqlQuery.ShowIndexes(TableName: ShortString);
begin
  if Active then Close;
  SQL := 'SELECT RDB$INDEX_ID AS Idx, A.RDB$INDEX_NAME AS Name, RDB$RELATION_NAME'
    +' AS Tbl, RDB$UNIQUE_FLAG AS Uni, RDB$INDEX_TYPE AS Srt, RDB$FIELD_NAME AS Fld'
    +' FROM RDB$INDICES A LEFT JOIN RDB$INDEX_SEGMENTS B'
    +' ON A.RDB$INDEX_NAME=B.RDB$INDEX_NAME WHERE';
  if TableName <> '' then
    SQL := SQL + ' RDB$RELATION_NAME LIKE '''+ UpperCase(TableName)+' %'''
  else
    SQL := SQL + ' RDB$INDEX_NAME NOT LIKE ''RDB$%''';
  Open;
end;

{ Showes tables of database }
procedure TDirIbSqlQuery.ShowTables(TableName: ShortString);
begin
  if Active then Close;
  Sql := 'SELECT RDB$RELATION_ID AS Idx, RDB$RELATION_NAME AS TableName'
    +' FROM RDB$RELATIONS WHERE';
  if TableName <> '' then
    Sql := Sql + ' RDB$RELATION_NAME LIKE '''+UpperCase(TableName)+''''
  else
    Sql := Sql + ' RDB$RELATION_NAME NOT LIKE ''RDB$%''';
  Sql := Sql + ' ORDER BY RDB$RELATION_NAME';
  Open;
end;

{ Convert string to sql format }
function TDirIbSqlQuery.StringToSql(Value: string): string;
begin
  Result := Value;
end;

{*************** TDirIbSqlBlob implementation ****************}

{ Class constructor }
constructor TDirIbSqlBlob.Create(AConnect: TDirConnect; ATransact: TDirTransact;
  AHandle: TBlobHandle);
begin
  inherited Create(AConnect, ATransact, AHandle);
end;

{ Open a blob }
procedure TDirIbSqlBlob.Open(Mode: Integer);
var
  IbConnect: TDirIbSqlConnect;
  IbTransact: TDirIbSqlTransact;
begin
  SetStatus(bsFail);
  if not Assigned(Connect) or not Assigned(Transact) then
    Exit;
  if not Connect.Active or not Transact.Active then
    Exit;

  if Active then Close;
  IbConnect := TDirIbSqlConnect(Connect);
  IbTransact := TDirIbSqlTransact(Transact);

  if Mode = fmOpenRead then
  begin
    FBlobHandle := nil;
    isc_open_blob2(@IbConnect.FStatusVector, @IbConnect.Handle, @IbTransact.Handle,
      @FBlobHandle, @Handle, 0, nil);
    if not ((IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0)) then
    begin
      SetStatus(bsOk);
      SetActive(True);
    end;
  end
  else if Mode = fmOpenWrite then
    CreateBlob;
end;

{ Close current blob }
procedure TDirIbSqlBlob.Close;
var
  IbConnect: TDirIbSqlConnect;
begin
  SetStatus(bsFail);
  if not Assigned(Connect) or not Connect.Active then
    Exit;

  if Active then
  begin
    IbConnect := TDirIbSqlConnect(Connect);
    isc_close_blob(@IbConnect.FStatusVector, @FBlobHandle);
    if not ((IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0)) then
      SetStatus(bsOk);
  end;
  SetActive(False);
end;

{ Create a new blob }
procedure TDirIbSqlBlob.CreateBlob;
var
  IbConnect: TDirIbSqlConnect;
  IbTransact: TDirIbSqlTransact;
begin
  if Active then Close;

  IbConnect := TDirIbSqlConnect(Connect);
  IbTransact := TDirIbSqlTransact(Transact);

  FBlobHandle := nil;
  isc_create_blob2(@IbConnect.FStatusVector, @IbConnect.Handle, @IbTransact.Handle,
    @FBlobHandle, @Handle, 0, nil);
  if not ((IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0)) then
  begin
    SetStatus(bsOk);
    SetActive(True);
  end;
end;

{ Delete current blob }
procedure TDirIbSqlBlob.DropBlob;
begin
  inherited DropBlob;
  if Assigned(Connect) and (FBlobHandle <> nil) then
    isc_cancel_blob(@TDirIbSqlConnect(Connect).FStatusVector, @FBlobHandle);
  SetStatus(bsOk);
  SetActive(False);
  PBlobHandle(@Handle).Ptr := 0;
  PBlobHandle(@Handle).PtrEx := 0;
  FBlobHandle := nil;
end;

{ Read segment from open blob }
function TDirIbSqlBlob.Read(Buffer: PChar; Length: Integer): Integer;
var
  TempLen: Short;
  IbConnect: TDirIbSqlConnect;
  BlobStat: LongInt;
begin
  Result := 0;
  SetStatus(bsFail);
  if not Assigned(Connect) or not Assigned(Transact) then
    Exit;
  if not Connect.Active or not Transact.Active or not Active then
    Exit;

  IbConnect := TDirIbSqlConnect(Connect);
  repeat
    BlobStat := isc_get_segment(@IbConnect.FStatusVector, @FBlobHandle,
      @TempLen, Length - Result, Buffer + Result);
    Inc(Result, TempLen);
  until (BlobStat = 335544367) or (Result >= Length);

  if not ((IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0)) then
    SetStatus(bsOk);
end;

{ Write segment to open blob }
function TDirIbSqlBlob.Write(Buffer: PChar; Length: Integer): Integer;
var
  IbConnect: TDirIbSqlConnect;
begin
  Result := 0;
  SetStatus(bsFail);

  if ((Handle.Ptr = 0) and (Handle.PtrEx = 0)) or (FBlobHandle = nil) then
    Exit;
  if not Assigned(Connect) or not Assigned(Transact) then
    Exit;
  if not Connect.Active or not Transact.Active or not Active then
    Exit;

  IbConnect := TDirIbSqlConnect(Connect);
  isc_put_segment(@IbConnect.FStatusVector, @FBlobHandle, Length, Buffer);
  if not ((IbConnect.StatusVector[0] = 1) and (IbConnect.StatusVector[1] > 0)) then
  begin
    SetStatus(bsOk);
    Result := Length;
  end;
end;

{**************** Extra functions *******************}

{ Convert postgresql field types to delphi field types }
function IbSqlToDelphiType(Value, Precision: Integer): TFieldType;
begin
  case Value of
    7: Result := ftSmallInt;
    8: if Precision <> 0 then Result := ftFloat
       else Result := ftInteger;
    10, 11, 27: Result := ftFloat;
    14, 37: Result := ftString;
    12: Result := ftDate;
    35: Result := ftDateTime;
    261: Result := ftBlob;
    else Result := ftUnknown;
  end;
end;

initialization
  MonitorList := TZMonitorList.Create;
finalization
  MonitorList.Free;
end.

