{********************************************************}
{                                                        }
{                 Zeos Database Objects                  }
{               Lexical Sql-Scanner class                }
{                                                        }
{       Copyright (c) 1999-2001 Sergey Seroukhov         }
{                                                        }
{********************************************************}

unit ZSqlScanner;

interface

uses Classes, SysUtils, ZScanner, ZSqlTypes;

type
{ Sql-scanner class definition }
  TZSqlScanner = class (TZScanner)
  private
    FDatabaseType: TDatabaseType;

    function MySqlLowRunLex(var CurrPos, CurrLineNo: Integer;
      var CurrToken: string): Integer;
    function IbSqlLowRunLex(var CurrPos, CurrLineNo: Integer;
      var CurrToken: string): Integer;
    function PgSqlLowRunLex(var CurrPos, CurrLineNo: Integer;
      var CurrToken: string): Integer;
    function MsSqlLowRunLex(var CurrPos, CurrLineNo: Integer;
      var CurrToken: string): Integer;


    function MySqlWrapString(Value: string): string;
    function IbSqlWrapString(Value: string): string;
    function PgSqlWrapString(Value: string): string;
    function MsSqlWrapString(Value: string): string;

    function MySqlUnwrapString(Value: string): string;
    function IbSqlUnwrapString(Value: string): string;
    function PgSqlUnwrapString(Value: string): string;
    function MsSqlUnwrapString(Value: string): string;
  protected
    function LowRunLex(var CurrPos, CurrLineNo: Integer;
      var CurrToken: string): Integer; override;

    function InnerProcSqlDelim(var Res: Integer; var Symb: Char;
      var CurrPos, CurrLineNo: Integer; var CurrToken: string): Boolean;
  public
    constructor Create; override;

    function WrapString(Value: string): string; override;
    function UnwrapString(Value: string): string; override;

    property DatabaseType: TDatabaseType read FDatabaseType write FDatabaseType;
  end;

implementation

{ TZSqlScanner }

{ Class constructor }
constructor TZSqlScanner.Create;
begin
  inherited Create;
  FDatabaseType := dtUnknown;
end;

{ Get lowlevel token }
function TZSqlScanner.LowRunLex(var CurrPos, CurrLineNo: Integer;
  var CurrToken: string): Integer;
begin
  case FDatabaseType of
    dtInterbase:  Result := IbSqlLowRunLex(CurrPos, CurrLineNo, CurrToken);
    dtMsSql:      Result := MsSqlLowRunLex(CurrPos, CurrLineNo, CurrToken);
    dtPostgreSql: Result := PgSqlLowRunLex(CurrPos, CurrLineNo, CurrToken);
    else          Result := MySqlLowRunLex(CurrPos, CurrLineNo, CurrToken);
  end;
end;

{ Unconvert value into string value }
function TZSqlScanner.UnwrapString(Value: string): string;
begin
  case FDatabaseType of
    dtInterbase:  Result := IbSqlUnwrapString(Value);
    dtMsSql:      Result := MsSqlUnwrapString(Value);
    dtPostgreSql: Result := PgSqlUnwrapString(Value);
    else          Result := MySqlUnwrapString(Value);
  end;
end;

{ Convert string value into string }
function TZSqlScanner.WrapString(Value: string): string;
begin
  case FDatabaseType of
    dtInterbase:  Result := IbSqlWrapString(Value);
    dtMsSql:      Result := MsSqlWrapString(Value);
    dtPostgreSql: Result := PgSqlWrapString(Value);
    else          Result := MySqlWrapString(Value);
  end;
end;

{ Process sql delimiters }
function TZSqlScanner.InnerProcSqlDelim(var Res: Integer; var Symb: Char;
  var CurrPos, CurrLineNo: Integer; var CurrToken: string): Boolean;
var
  Temp1: Char;
begin
  Result := True;
  { Check for brace }
  if Symb in ['{', '}', '(', ')', '[', ']'] then
  begin
    Res := tokBrace;
  end
  { Check for separator }
  else if Symb in [',', ';', ':'] then
  begin
    Res := tokSeparator;
  end
  { Check for delimiters }
  else if Pos(Symb, '~!#%?|=+-<>/*^@#') > 0 then
  begin
    Res := tokOperator;
    if FBufferPos <= FBufferLen then
      Temp1 := FBuffer[FBufferPos]
    else Temp1 := #0;
    if ((Symb = '>') and (Temp1 = '='))
      or ((Symb = '<') and (Temp1 in ['=', '>'])) then
    begin
      CurrToken := CurrToken + Temp1;
      Inc(FBufferPos);
    end;
  end
  else
    Result := False;
end;

const
  MaxIbSqlKeyword  = 24;

  IbSqlKeyword: array[1..MaxIbSqlKeyword] of string =
    ('select','update','delete','create','drop','explain','from','where',
      'order','by','group','having','set','values','into','union','all',
      'distinct','left','right','on','join','outer','inner');

{ Get lowlevel token for Interbase }
function TZSqlScanner.IbSqlLowRunLex(var CurrPos, CurrLineNo: Integer;
  var CurrToken: string): Integer;
var
  I: Integer;
  Temp: Char;
  Search: string;
begin
  Result := InnerStartLex(Temp, CurrPos, CurrLineNo, CurrToken);
  if FBufferPos > FBufferLen then Exit;
  { Check for EOL }
  if Temp = #13 then
  begin
    Result := tokEol;
    Inc(FBufferLine);
  end
  { Check for multi-line comment }
  else if InnerProcCComment(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    Result := tokComment;
  end
  { Check for single-line comment }
  else if (Temp = '/') and (FBufferPos <= FBufferLen)
    and (FBuffer[FBufferPos] = '/') then
  begin
    Result := InnerProcLComment(Temp, CurrPos, CurrLineNo, CurrToken);
  end
  { Check for sql delimiters }
  else if InnerProcSqlDelim(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    { }
  end
  { Check for string }
  else if InnerProcPasString(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    Result := tokString;
  end
  { Check for digits and identifiers }
  else if InnerProcIdent(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    { Check for keywords }
    if (Result = tokIdent) and ShowKeyword then
    begin
      Search := LowerCase(CurrToken);
      for I := 1 to MaxIbSqlKeyword do
        if IbSqlKeyword[I] = Search then
        begin
          Result := tokKeyword;
          Exit;
        end;
    end;
  end
  else
    Result := tokUnknown;
end;

{ Unconvert value into string value for Interbase }
function TZSqlScanner.IbSqlUnwrapString(Value: string): string;
var
  Pos, Len: Integer;
begin
  Result := '';
  if Value = '' then Exit;
  Pos := 1;
  Delete(Value, 1, 1);
  Len := Length(Value);
  while Pos <= Len do
  begin
    if not (Value[Pos] in ['''', '"']) then
      Result := Result + Value[Pos]
    else if (Pos < Len) and (Value[Pos+1] = Value[Pos]) then
    begin
      Result := Result + Value[Pos];
      Inc(Pos);
    end;
    Inc(Pos);
  end;
end;

{ Convert string value into string for Interbase }
function TZSqlScanner.IbSqlWrapString(Value: string): string;
var
  Pos: Integer;
begin
  Result := '''';
  for Pos := 1 to Length(Value) do
  begin
    Result := Result + Value[Pos];
    if Value[Pos] = '''' then
      Result := Result + Value[Pos];
  end;
  Result := Result + '''';
end;

{ Get lowlevel token for MS SQL }
function TZSqlScanner.MsSqlLowRunLex(var CurrPos, CurrLineNo: Integer;
  var CurrToken: string): Integer;
begin
  Result := IbSqlLowRunLex(CurrPos, CurrLineNo, CurrToken);
end;

{ Unconvert value into string value for MS SQL }
function TZSqlScanner.MsSqlUnwrapString(Value: string): string;
begin
  Result := IbSqlUnwrapString(Value);
end;

{ Convert string value into string for MS SQL }
function TZSqlScanner.MsSqlWrapString(Value: string): string;
begin
  Result := IbSqlWrapString(Value);
end;

{ Get lowlevel token for MySql }
function TZSqlScanner.MySqlLowRunLex(var CurrPos, CurrLineNo: Integer;
  var CurrToken: string): Integer;
var
  I: Integer;
  Temp: Char;
  Search: string;
begin
  Result := InnerStartLex(Temp, CurrPos, CurrLineNo, CurrToken);
  if FBufferPos > FBufferLen then Exit;
  { Check for EOL }
  if Temp = #13 then
  begin
    Result := tokEol;
    Inc(FBufferLine);
  end
  { Check for single-line comment }
  else if Temp = '#' then
  begin
    Result := InnerProcLComment(Temp, CurrPos, CurrLineNo, CurrToken);
  end
  { Check for sql delimiters }
  else if InnerProcSqlDelim(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    { }
  end
  { Check for string }
  else if InnerProcCString(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    Result := tokString;
  end
  { Check for digits and identifiers }
  else if InnerProcIdent(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    { Check for keywords }
    if (Result = tokIdent) and ShowKeyword then
    begin
      Search := LowerCase(CurrToken);
      for I := 1 to MaxIbSqlKeyword do
        if IbSqlKeyword[I] = Search then
        begin
          Result := tokKeyword;
          Exit;
        end;
    end;
  end
  else
    Result := tokUnknown;
end;

{ Unconvert value into string value for MySql }
function TZSqlScanner.MySqlUnwrapString(Value: string): string;
var
  N: Integer;
  Ptr1, Ptr2: PChar;
begin
  Result := '';
  if Value = '' then Exit;
  Delete(Value, 1, 1);
  Delete(Value, Length(Value), 1);

  SetLength(Result, Length(Value)+1);
  Ptr1 := PChar(Value);
  Ptr2 := PChar(Result);
  N := 0;
  while Ptr1^ <> #0 do
  begin
    if Ptr1^ <> '\' then
      Ptr2^ := Ptr1^
    else begin
      Inc(Ptr1);
      if Ptr1 = #0 then Break;
      case Ptr1^ of
        'n': Ptr2^ := #10;
        'r': Ptr2^ := #13;
        'Z': Ptr2^ := #26;
        '0': Ptr2^ := #0;                     
        else Ptr2^ := Ptr1^;
      end;
    end;
    Inc(N);
    Inc(Ptr1);
    Inc(Ptr2);
  end;
  SetLength(Result, N);
end;

{ Convert string value into string for MySql }
function TZSqlScanner.MySqlWrapString(Value: string): string;
var
  I, Add, Len: Integer;
  Ptr: PChar;
begin
  Add := 0;
  Len := Length(Value);
  for I := 1 to Len do
    if Value[I] in ['''','"','\',#26,#10,#13,#0] then
      Inc(Add);
  SetLength(Result, Len + Add);
  Ptr := PChar(Result);
  for I := 1 to Len do
  begin
    if Value[I] in ['''','"','\',#26,#10,#13,#0] then
    begin
      Ptr^ := '\';
      Inc(Ptr);
      case Value[I] of
        #26: Ptr^ := 'Z';
        #10: Ptr^ := 'n';
        #13: Ptr^ := 'r';
        #0:  Ptr^ := '0';
        else Ptr^ := Value[I];
      end;
    end else
      Ptr^ := Value[I];
    Inc(Ptr);
  end;
  Result := '''' + Result + '''';
end;

{ Get lowlevel token for PostgreSql }
function TZSqlScanner.PgSqlLowRunLex(var CurrPos, CurrLineNo: Integer;
  var CurrToken: string): Integer;
var
  I: Integer;
  Temp: Char;
  Search: string;
begin
  Result := InnerStartLex(Temp, CurrPos, CurrLineNo, CurrToken);
  if FBufferPos > FBufferLen then Exit;
  { Check for EOL }
  if Temp = #13 then
  begin
    Result := tokEol;
    Inc(FBufferLine);
  end
  { Check for multi-line comment }
  else if InnerProcCComment(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    Result := tokComment;
  end
  { Check for single-line comment }
  else if (Temp = '-') and (FBufferPos <= FBufferLen)
    and (FBuffer[FBufferPos] = '-') then
  begin
    Result := InnerProcLComment(Temp, CurrPos, CurrLineNo, CurrToken);
  end
  { Check for sql delimiters }
  else if InnerProcSqlDelim(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    { }
  end
  { Check for string }
  else if InnerProcCString(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    Result := tokString;
  end
  { Check for digits and identifiers }
  else if InnerProcIdent(Result, Temp, CurrPos, CurrLineNo, CurrToken) then
  begin
    { Check for keywords }
    if (Result = tokIdent) and ShowKeyword then
    begin
      Search := LowerCase(CurrToken);
      for I := 1 to MaxIbSqlKeyword do
        if IbSqlKeyword[I] = Search then
        begin
          Result := tokKeyword;
          Exit;
        end;
    end;
  end
  else
    Result := tokUnknown;
end;

{ Unconvert value into string value for PostgreSql }
function TZSqlScanner.PgSqlUnwrapString(Value: string): string;
var
  N: Integer;
  Ptr1, Ptr2: PChar;
begin
  Result := '';
  if Value = '' then Exit;
  Delete(Value, 1, 1);
  Delete(Value, Length(Value), 1);

  SetLength(Result, Length(Value)+1);
  Ptr1 := PChar(Value);
  Ptr2 := PChar(Result);
  N := 0;
  while Ptr1^ <> #0 do
  begin
    if Ptr1^ <> '\' then
      Ptr2^ := Ptr1^
    else begin
      Inc(Ptr1);
      if Ptr1 = #0 then Break;
      case Ptr1^ of
        'n': Ptr2^ := #10;
        'r': Ptr2^ := #13;
        't': Ptr2^ := #9;
        '0': Ptr2^ := #0;
        else Ptr2^ := Ptr1^;
      end;
    end;
    Inc(N);
    Inc(Ptr1);
    Inc(Ptr2);
  end;
  SetLength(Result, N);
end;

{ Convert string value into string for PostgreSql }
function TZSqlScanner.PgSqlWrapString(Value: string): string;
var
  I, Add, Len: Integer;
  Ptr: PChar;
begin
  Add := 0;
  Len := Length(Value);
  for I := 1 to Len do
    if Value[I] in ['''','"','\',#9,#10,#13,#0] then
      Inc(Add);
  SetLength(Result, Len + Add);
  Ptr := PChar(Result);
  for I := 1 to Len do
  begin
    if Value[I] in ['''','"','\',#9,#10,#13,#0] then
    begin
      Ptr^ := '\';
      Inc(Ptr);
      case Value[I] of
        #9:  Ptr^ := 't';
        #10: Ptr^ := 'n';
        #13: Ptr^ := 'r';
        #0:  Ptr^ := '0';
        else Ptr^ := Value[I];
      end;
    end else
      Ptr^ := Value[I];
    Inc(Ptr);
  end;
  Result := '''' + Result + '''';
end;

end.
